import torch
import numpy as np
from torch.utils.data import IterableDataset, get_worker_info

import pyarrow.parquet as pq
import orjson

from ochat.training_deepspeed.multipack_sampler import MultipackDistributedSampler


def _find_multiple(a, b):
    return (-(a // -b)) * b


class OpenchatDataset(IterableDataset):
    def __init__(self, dataset_filename, batch_max_length, rank, num_replicas):
        super().__init__()
        # Init constants
        self.PAD_ID = 0
        self.PAD_MULTIPLE = 64
        self.BATCH_KEYS = {
            "seqlens": torch.int32,
            "nz_input_ids": torch.long,
            "nz_position_ids": torch.long,
            "nz_shifted_label_ids": torch.long,

            "nz_shifted_loss_weights": torch.bfloat16
        }

        assert batch_max_length % self.PAD_MULTIPLE == 0, f"Batch size {batch_max_length} need to be multiples of {self.PAD_MULTIPLE}"

        # Load data
        # Convert parquet to numpy for fast random access
        table = pq.read_table(dataset_filename, memory_map=True)
        self.dataset = {k: v.to_numpy() for k, v in zip(table.column_names, table.columns)}

        # read metadata
        self.metadata = table.schema.metadata.get(b"metadata_json", None)
        if self.metadata is not None:
            self.metadata = orjson.loads(self.metadata)

        # Free table space
        del table

        # Create sampler
        self.sampler = MultipackDistributedSampler(
            lengths=self.dataset["total_length"],
            numseqs=self.dataset["num_seqs"],

            batch_max_length=batch_max_length,

            rank=rank,
            num_replicas=num_replicas,
            seed=0
        )

        # Init state
        self._epoch = 0

    def _load_batch(self, indices):
        batch = {k: v[indices] for k, v in self.dataset.items()}

        # Concat batches
        batch = {k: np.concatenate(batch[k], axis=0) for k in self.BATCH_KEYS.keys()}

        # Pad an unused item to reach multiple of PAD_MULTIPLE, for faster GEMM
        total_seqlen = batch["nz_input_ids"].size
        pad_len      = _find_multiple(total_seqlen, self.PAD_MULTIPLE) - total_seqlen

        if pad_len > 0:
            assert pad_len < self.PAD_MULTIPLE

            # total length
            padding_specs = {
                "seqlens": (1, pad_len),

                "nz_input_ids": (pad_len, self.PAD_ID),
                "nz_position_ids": (pad_len, 0),
                "nz_shifted_label_ids": (pad_len, self.PAD_ID),
                "nz_shifted_loss_weights": (pad_len, 0),
            }
            for k, pad_spec in padding_specs.items():
                batch[k] = np.concatenate((batch[k], np.full(*pad_spec, dtype=batch[k].dtype)), axis=0)

        # to tensor
        batch_tensor = {}
        for k, dtype in self.BATCH_KEYS.items():
            batch_tensor[k] = torch.from_numpy(batch[k]).to(dtype)

        # cu seqlens
        batch_tensor["cu_seqlens"] = torch.nn.functional.pad(batch_tensor["seqlens"].cumsum(-1, dtype=torch.int32), (1, 0))
        # batch info
        batch_info = {"max_seqlen": torch.max(batch_tensor["seqlens"]).item()}

        # inputs
        del batch_tensor["seqlens"]
        return batch_tensor, batch_info

    def __iter__(self):
        worker_info = get_worker_info()
        assert worker_info is None or worker_info.num_workers == 1

        for indices, all_numseq, cur_numseq in self.sampler.iter(self._epoch):
            yield self._load_batch(indices), all_numseq, cur_numseq

        # Increase epoch count
        self._epoch += 1

    def estimate_num_batches(self):
        return self.sampler.estimate_num_batches()



import numpy as np
import numba


@numba.njit
def ffd_check(a: np.ndarray, c: int, n: int):
    # First-fit-decreasing bin packing
    # Check if a[] could fit in n bins with capacity c
    # https://en.wikipedia.org/wiki/First-fit-decreasing_bin_packing

    a = np.sort(a)[::-1]
    bins = np.full((n, ), c, dtype=a.dtype)
    for size in a:
        not_found = True
        for idx in range(n):
            if bins[idx] >= size:
                bins[idx] -= size
                not_found = False
                break

        if not_found:
            return False

    return True


@numba.njit
def ffd_with_result(a: np.ndarray, c: int, start_index: int):
    # First-fit-decreasing bin packing (with result return)

    indices = np.argsort(a)[::-1]
    a = a[indices]

    bins = []
    bins_result = []
    for a_id, size in enumerate(a):
        add_new = True
        for idx in range(len(bins)):
            if bins[idx] >= size:
                bins[idx] -= size
                bins_result[idx].append(indices[a_id] + start_index)
                add_new = False
                break

        if add_new:
            bins.append(c - size)
            bins_result.append([indices[a_id] + start_index])

    return bins_result


@numba.njit
def allocate(lengths: np.ndarray, numseqs: np.ndarray, lengths_cumsum: np.ndarray, rank: int, c: int, n: int):
    # Dynamic batch allocator, similar to Multifit
    # https://en.wikipedia.org/wiki/Multifit_algorithm
    # ~99.5% efficiency on OpenChat training set (12 * 2048 ctx len)

    s = 0
    start_index = 0
    result = []
    result_totseqs = []

    while True:
        # binary search [l, r)
        l = 1
        r = 1 + np.searchsorted(lengths_cumsum[start_index:], s + c * n, "right")

        while r - l > 1:
            m = (l + r) // 2
            if ffd_check(lengths[start_index: start_index + m], c, n):
                l = m
            else:
                r = m

        # use length l
        batch = ffd_with_result(lengths[start_index: start_index + l], c, start_index)  # type: ignore
        if len(batch) < n:
            break

        start_index += l
        s = lengths_cumsum[start_index - 1]

        # add local rank
        result.append(batch[rank])
        # add total seqs for all ranks
        totseq = 0
        for indices in batch:
            for idx in indices:
                totseq += numseqs[idx]
        result_totseqs.append(totseq)

    return result, result_totseqs, s, len(result) * c * n


class MultipackDistributedSampler:
    """Unpadded data loading using Multipack.
       Approximate (at most ~1.22x) the optimal solution of the identical-machines scheduling problem, which is NP-hard."""
    
    def __init__(
        self,
        lengths: np.ndarray,
        numseqs: np.ndarray,

        batch_max_length: int,

        num_replicas: int,
        rank: int,

        seed: int,
    ):
        # Dataset
        self.lengths = lengths
        self.numseqs = numseqs
        assert isinstance(self.lengths, np.ndarray)

        self.batch_max_length = batch_max_length

        # Get rank
        self.num_replicas = num_replicas
        self.rank = rank

        # Seed
        self.seed = seed

        # statistics
        self.eff_total_used = 0
        self.eff_total_slots = 0

    def generate_batches(self, epoch, set_stats=False):
        indices = np.random.default_rng(seed=self.seed + epoch).permutation(len(self.lengths))

        lengths        = self.lengths[indices]
        numseqs        = self.numseqs[indices]
        lengths_cumsum = np.cumsum(lengths)

        batches, totseqs, total_used, total_slots = allocate(lengths=lengths,
                                                             numseqs=numseqs,
                                                             lengths_cumsum=lengths_cumsum,
                                                             rank=self.rank,
                                                             c=self.batch_max_length,
                                                             n=self.num_replicas)
        
        curseqs = [np.sum(numseqs[batch]) for batch in batches]
        batches = [indices[batch]         for batch in batches]

        # statistics
        if set_stats:
            self.eff_total_used += total_used
            self.eff_total_slots += total_slots

        return batches, totseqs, curseqs
    
    def iter(self, epoch):
        all_batches, all_totseqs, all_curseqs = self.generate_batches(epoch, set_stats=True)

        for batch, totseq, curseq in zip(all_batches, all_totseqs, all_curseqs):
            yield batch, totseq, curseq

    def estimate_num_batches(self):
        batches, _, _ = self.generate_batches(epoch=0)
        return len(batches)

    def efficiency(self):
        return self.eff_total_used / self.eff_total_slots



from typing import Optional
import argparse
import os
import asyncio
from glob import glob

import orjson
import openai
from tqdm import tqdm
from openai import RateLimitError, InternalServerError, APIConnectionError
from tenacity import retry, stop_after_attempt, wait_random_exponential, retry_if_exception_type
from vllm import LLM, SamplingParams

from transformers.utils.hub import cached_file

from ochat.evaluation.match_answer import MATCH_ANSWER_FUNCTION
from ochat.config import MODEL_CONFIG_MAP


def _strip_first_space(s: str):
    if len(s) and s[0] == " ":
        return s[1:]
    return s


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(20), retry=retry_if_exception_type((RateLimitError, InternalServerError, APIConnectionError, )))
async def _chat_completion_with_backoff(client, **kwargs):
    return await client.chat.completions.create(**kwargs)


async def chat_completion_thread(model, progress_bar, queue):
    client = openai.AsyncOpenAI()

    while True:
        # Fetch task
        try:
            task = queue.get_nowait()
        except asyncio.QueueEmpty:
            break

        # Completion
        try:
            response = await _chat_completion_with_backoff(
                client,
                model=model,
                messages=[{"role": "user", "content": task["question"]}],

                temperature=0
            )
            task["response"] = response.choices[0].message.content  # type: ignore
        except Exception as e:
            if hasattr(e, "last_attempt"):
                e = e.last_attempt
            if hasattr(e, "_exception"):
                e = e._exception

            print(type(e), str(e))
        
        # Progress
        progress_bar.update()


async def get_openai_answers(
    model: str,
    questions: list,
    parallel: int
):
    # Complete in retry cycles
    last_to_complete_num = 0

    while True:
        # fill queue
        to_complete_num = 0
        queue = asyncio.Queue()
        for q in questions:
            if q["response"]:
                continue

            queue.put_nowait(q)
            to_complete_num += 1

        tqdm.write(f"New completion cycle. To complete {to_complete_num}, number of parallel calls {parallel}")

        # Create tasks
        progress_bar = tqdm(total=to_complete_num)
        async with asyncio.TaskGroup() as task_group:
            for _ in range(parallel):
                task_group.create_task(chat_completion_thread(model, progress_bar, queue))

        # Next retry cycle
        # Break if cannot complete more
        if (to_complete_num == last_to_complete_num) or (to_complete_num == 0):
            break
        last_to_complete_num = to_complete_num

        # Reduce parallel calls
        parallel = max(1, parallel // 2)

    return questions


def tokenize_questions(model_config: object, conv_template: object, questions: list, condition: str, system_msg: str):
    from ochat.config import Conversation, Message

    # Construct conversation
    prompt_indices = []
    conversations = []
    for idx, q in enumerate(questions):
        if q["response"]:
            continue

        conversations.append(Conversation(
            items=[
                Message(role="user", content=q["question"]),
                Message(role="assistant", content="")
            ],
            condition=condition,
            system=system_msg
        ))
        prompt_indices.append(idx)

    # Tokenize
    conversations, _ = conv_template.tokenize_conversations(conversations, inference=True)
    conversations    = [tokens[-model_config.model_max_context:] for tokens in conversations]

    return conversations, prompt_indices


def get_model_answers(
    model: str,
    questions: list,
    condition: str,
    system_msg: str,
    model_type: str,
    tensor_parallel_size: int
):
    # Load model config
    if model_type is None:
        with open(cached_file(path_or_repo_id=model, filename="openchat.json"), "r") as f:
            model_type = orjson.loads(f.read())["model_type"]

    model_config = MODEL_CONFIG_MAP[model_type]
    tokenizer = model_config.model_tokenizer_create(model)
    conv_template = model_config.conversation_template(tokenizer=tokenizer)

    # Init vLLM engine
    engine = LLM(model,
                 max_num_batched_tokens=model_config.model_max_context,
                 max_model_len=model_config.model_max_context,
                 tensor_parallel_size=tensor_parallel_size)
    sampling_params = SamplingParams(temperature=0,
                                     max_tokens=None,
                                     stop_token_ids=conv_template.eot_tokens_,  # Override stop tokens
                                     ignore_eos=True)

    # Complete
    prompts, prompt_indices = tokenize_questions(model_config, conv_template, questions,
                                                 condition=condition, system_msg=system_msg)

    # calculate & fill in responses
    responses = engine.generate(prompt_token_ids=prompts, sampling_params=sampling_params)
    for idx, resp in zip(prompt_indices, responses):
        questions[idx]["response"] = _strip_first_space(resp.outputs[0].text)

    return questions


async def run_eval(
    model: str,
    condition: str,
    system_msg: str,
    model_type: str,

    data_path: str,
    eval_sets: list,

    continue_from: Optional[str],
    output_file: str,

    parallel: int,
    tensor_parallel_size: int
):
    print (f"Evaluating ({model_type})...\n\nCondition: {condition}\nSystem Prompt: {system_msg}\n")

    if continue_from is not None:
        # Load continue
        print (f"Continuing from {continue_from}...")

        with open(continue_from, "rb") as f:
            questions = orjson.loads(f.read())
    else:
        # Load questions
        questions = []

        for filename in glob(os.path.join(data_path, "**", "*.jsonl"), recursive=True):
            task_name = os.path.splitext(filename[len(data_path):])[0].strip("\\/")
            task_type = os.path.dirname(task_name)

            assert task_type in MATCH_ANSWER_FUNCTION

            # Filter eval sets
            if eval_sets and not sum([task_name.startswith(a) for a in eval_sets]):
                continue

            # Load task
            with open(filename, "r") as f:
                task_data = list(map(orjson.loads, f.readlines()))

            questions.extend([{**item, "task_name": task_name, "task_type": task_type, "response": ""} for item in task_data])

    # run completion
    if model.startswith("gpt-3.5-turbo") or model.startswith("gpt-4"):
        questions = await get_openai_answers(model, questions, parallel)
    else:
        questions = get_model_answers(model, questions, condition, system_msg, model_type, tensor_parallel_size)

    # Calculate accuracy
    for q in questions:
        q["is_matched"], q["answer"] = MATCH_ANSWER_FUNCTION[q["task_type"]](q, q["response"])
        try:
            q["is_correct"] = q["answer"] in q["label"]
        except:
            q["is_correct"] = False

    # Write results
    if output_file is None:
        output_file = os.path.join(os.path.dirname(data_path), "eval_results", f"{os.path.basename(model)}_{condition}.json")

    os.makedirs(os.path.dirname(output_file), exist_ok=True)
    with open(output_file, "wb") as f:
        f.write(orjson.dumps(questions, option=orjson.OPT_INDENT_2))


async def main():
    parser = argparse.ArgumentParser()

    # Input / output
    parser.add_argument("--model", type=str, default=None)
    parser.add_argument("--condition", type=str, default="")
    parser.add_argument("--system-msg", type=str, default="")
    parser.add_argument("--model-type", type=str, default=None)

    parser.add_argument("--data-path", type=str, default="ochat/evaluation/eval_data")
    parser.add_argument("--eval-sets", type=str, nargs="+", default=[])

    parser.add_argument("--continue-from", type=str, default=None)
    parser.add_argument("--output-file",   type=str, default=None)
    parser.add_argument("--parallel",      type=int, default=16)
    parser.add_argument("--tensor-parallel-size", type=int, default=1)

    args = parser.parse_args()

    await run_eval(**vars(args))

if __name__ == "__main__":
    asyncio.run(main())



# Adapted from
# https://github.com/vllm-project/vllm/blob/main/vllm/entrypoints/openai/api_server.py

import argparse
import asyncio
from http import HTTPStatus
import json
import time
import logging
from logging.handlers import RotatingFileHandler
from typing import AsyncGenerator, Optional
from dataclasses import dataclass

import fastapi
from fastapi import BackgroundTasks, Request
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, StreamingResponse
from fastapi.security.http import HTTPAuthorizationCredentials, HTTPBearer

import uvicorn
import ray

from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.outputs import RequestOutput
from vllm.sampling_params import SamplingParams
from vllm.utils import random_uuid

from ochat.config import MODEL_CONFIG_MAP
from ochat.serving import openai_api_protocol, async_tokenizer

from transformers.utils.hub import cached_file


TIMEOUT_KEEP_ALIVE = 5  # seconds


@dataclass
class ModelConfig:
    names: set = None

    max_length: int = None
    stream_period: int = None
    eot_tokens: list = None

    enable_sys_prompt: bool = None
    api_keys: list = None


logger = None
app = fastapi.FastAPI()

model = ModelConfig()
tokenizer = None


def _strip_first_space(s: str):
    if s[0] == " ":
        return s[1:]
    return s


def log_request(created_time: int, request: openai_api_protocol.ChatCompletionRequest, output: RequestOutput):
    if logger is not None:
        logger.info(openai_api_protocol.LoggingRecord(
            time=created_time,
            request=request,
            outputs=[o.text for o in output.outputs]
        ).model_dump_json(exclude_unset=True))


def create_error_response(status_code: HTTPStatus,
                          message: str) -> JSONResponse:
    return JSONResponse(openai_api_protocol.ErrorResponse(message=message,
                                                          type="invalid_request_error").dict(),
                        status_code=status_code.value)


def check_model(request) -> Optional[JSONResponse]:
    if request.model in model.names:
        return

    return create_error_response(
        HTTPStatus.NOT_FOUND,
        f"The model `{request.model}` does not exist.",
    )


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):  # pylint: disable=unused-argument
    return create_error_response(HTTPStatus.BAD_REQUEST, str(exc))


async def check_api_key(
    auth: Optional[HTTPAuthorizationCredentials] = fastapi.Depends(HTTPBearer(auto_error=False)),
):
    if not model.api_keys:
        return

    if auth is None or auth.credentials not in model.api_keys:
        raise fastapi.HTTPException(
            status_code=401,
            detail={
                "error": {
                    "message": "",
                    "type": "invalid_request_error",
                    "param": None,
                    "code": "invalid_api_key",
                }
            },
        )


@app.get("/v1/models", dependencies=[fastapi.Depends(check_api_key)])
async def show_available_models():
    """Show available models. Right now we only have one model."""
    return openai_api_protocol.ModelList(data=[
        openai_api_protocol.ModelCard(id=name,
                                      root=name,
                                      permission=[openai_api_protocol.ModelPermission()])
    for name in model.names])


@app.post("/v1/chat/completions", dependencies=[fastapi.Depends(check_api_key)])
async def create_chat_completion(raw_request: Request, background_tasks: BackgroundTasks):
    """Completion API similar to OpenAI's API.

    See  https://platform.openai.com/docs/api-reference/chat/create
    for the API specification. This API mimics the OpenAI ChatCompletion API.

    NOTE: Currently we do not support the following features:
        - function_call (Users should implement this by themselves)
        - logit_bias (to be supported by vLLM engine)
    """

    request = openai_api_protocol.ChatCompletionRequest(**await raw_request.json())

    error_check_ret = check_model(request)
    if error_check_ret is not None:
        return error_check_ret

    if request.logit_bias is not None and len(request.logit_bias) > 0:
        # TODO: support logit_bias in vLLM engine.
        return create_error_response(HTTPStatus.BAD_REQUEST,
                                     "logit_bias is not currently supported")

    # input ids
    input_ids = await tokenizer.tokenize.remote(request.messages, condition=request.condition,
                                                enable_sys_prompt=model.enable_sys_prompt)
    input_num_tokens = len(input_ids)

    # check length
    if request.max_tokens is None:
        request.max_tokens = model.max_length - input_num_tokens

    if input_num_tokens + request.max_tokens > model.max_length:
        return input_ids, create_error_response(
            HTTPStatus.BAD_REQUEST,
            f"This model's maximum context length is {model.max_length} tokens. "
            f"However, you requested {input_num_tokens + request.max_tokens} tokens "
            f"({input_num_tokens} in the messages, "
            f"{request.max_tokens} in the completion). "
            f"Please reduce the length of the messages or completion.",
        )

    # completion
    model_name = request.model
    request_id = f"cmpl-{random_uuid()}"
    created_time = int(time.time())

    try:
        sampling_params = SamplingParams(
            n=request.n,
            presence_penalty=request.presence_penalty,
            frequency_penalty=request.frequency_penalty,
            temperature=request.temperature,
            top_p=request.top_p,
            max_tokens=request.max_tokens,
            seed=request.seed,
            # Override stop tokens
            stop_token_ids=model.eot_tokens,
            ignore_eos=True
        )
    except ValueError as e:
        return create_error_response(HTTPStatus.BAD_REQUEST, str(e))

    result_generator = engine.generate(prompt=None,
                                       prompt_token_ids=input_ids,
                                       sampling_params=sampling_params,
                                       request_id=request_id)

    def create_stream_response_json(
        index: int,
        text: str,
        finish_reason: Optional[str] = None,
    ) -> str:
        choice_data = openai_api_protocol.ChatCompletionResponseStreamChoice(
            index=index,
            delta=openai_api_protocol.DeltaMessage(content=text),
            finish_reason=finish_reason,
        )
        response = openai_api_protocol.ChatCompletionStreamResponse(
            id=request_id,
            choices=[choice_data],
            model=model_name,
        )

        return response.model_dump_json(exclude_unset=True)

    async def completion_stream_generator() -> AsyncGenerator[str, None]:
        # First chunk with role
        for i in range(request.n):
            choice_data = openai_api_protocol.ChatCompletionResponseStreamChoice(
                index=i,
                delta=openai_api_protocol.DeltaMessage(role="assistant"),
                finish_reason=None,
            )
            chunk = openai_api_protocol.ChatCompletionStreamResponse(id=request_id,
                                                                     choices=[choice_data],
                                                                     model=model_name)

            yield f"data: {chunk.model_dump_json(exclude_unset=True)}\n\n"

        previous_texts = [""] * request.n
        previous_num_tokens = [0] * request.n

        stream_index = 0
        final_res = None
        is_first = True
        async for res in result_generator:
            stream_index += 1
            final_res = res

            for output in res.outputs:
                # stream on end or every stream_period
                if (stream_index % model.stream_period == 0) or (output.finish_reason is not None):
                    i = output.index
                    delta_text = output.text[len(previous_texts[i]):]
                    if "\ufffd" not in delta_text:
                        previous_texts[i] = output.text
                        previous_num_tokens[i] = len(output.token_ids)

                        if is_first:
                            # Strip first space
                            is_first = False
                            delta_text = _strip_first_space(delta_text)

                        yield f"data: {create_stream_response_json(index=i, text=delta_text)}\n\n"
                        if output.finish_reason is not None:
                            yield f"data: {create_stream_response_json(index=i, text='', finish_reason=output.finish_reason)}\n\n"

        yield "data: [DONE]\n\n"

        # Log request
        background_tasks.add_task(log_request, created_time, request, final_res)

    # Streaming response
    if request.stream:
        return StreamingResponse(completion_stream_generator(),
                                 media_type="text/event-stream")

    # Non-streaming response
    final_res: RequestOutput = None
    async for res in result_generator:
        if await raw_request.is_disconnected():
            # Abort the request if the client disconnects.
            await engine.abort(request_id)
            return create_error_response(HTTPStatus.BAD_REQUEST,
                                         "Client disconnected")
        final_res = res
    assert final_res is not None
    choices = []
    for output in final_res.outputs:
        choice_data = openai_api_protocol.ChatCompletionResponseChoice(
            index=output.index,
            message=openai_api_protocol.ChatMessage(role="assistant", content=_strip_first_space(output.text)),
            finish_reason=output.finish_reason,
        )
        choices.append(choice_data)

    num_prompt_tokens = len(final_res.prompt_token_ids)
    num_generated_tokens = sum(
        len(output.token_ids) for output in final_res.outputs)
    usage = openai_api_protocol.UsageInfo(
        prompt_tokens=num_prompt_tokens,
        completion_tokens=num_generated_tokens,
        total_tokens=num_prompt_tokens + num_generated_tokens,
    )
    response = openai_api_protocol.ChatCompletionResponse(
        id=request_id,
        created=created_time,
        model=model_name,
        choices=choices,
        usage=usage,
    )

    # Log request
    background_tasks.add_task(log_request, created_time, request, final_res)

    return response


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="OpenChat OpenAI-Compatible RESTful API server.")

    # Model
    parser.add_argument("--model-type", type=str, default=None, help="Model type. Leave empty to auto-detect.")

    parser.add_argument("--stream-period", type=int, default=6, help="Number of tokens per stream event")
    parser.add_argument("--api-keys", type=str, nargs="*", default=[], help="Allowed API Keys. Leave blank to not verify")
    parser.add_argument("--enable-sys-prompt", default=False, action="store_true")

    # Server
    parser.add_argument("--host", type=str, default="localhost", help="Host name")
    parser.add_argument("--port", type=int, default=18888, help="Port number")
    parser.add_argument("--allow-credentials", action="store_true", help="Allow credentials")
    parser.add_argument("--allowed-origins", type=json.loads, default=["*"], help="Allowed origins")
    parser.add_argument("--allowed-methods", type=json.loads, default=["*"], help="Allowed methods")
    parser.add_argument("--allowed-headers", type=json.loads, default=["*"], help="Allowed headers")

    # Logging
    parser.add_argument("--log-file", type=str, default=None, help="Log file. Leave blank to disable logging")
    parser.add_argument("--log-max-mb", type=int, default=128, help="Max log size in MB")
    parser.add_argument("--log-max-count", type=int, default=10, help="Max log file versions to keep")

    parser = AsyncEngineArgs.add_cli_args(parser)
    args = parser.parse_args()

    # App and logging
    app.add_middleware(
        CORSMiddleware,
        allow_origins=args.allowed_origins,
        allow_credentials=args.allow_credentials,
        allow_methods=args.allowed_methods,
        allow_headers=args.allowed_headers,
    )

    if args.log_file:
        logger = logging.getLogger(__name__)

        logger.setLevel(logging.INFO)
        logger.addHandler(RotatingFileHandler(
            args.log_file,
            maxBytes=args.log_max_mb * 1048576,
            backupCount=args.log_max_count)
        )
        logger.propagate = False

    # Load model type
    if args.model_type is None:
        with open(cached_file(path_or_repo_id=args.model, filename="openchat.json"), "r") as f:
            args.model_type = json.load(f)["model_type"]

    # Load tokenizer
    tokenizer = async_tokenizer.AsyncTokenizer.remote(args.model_type, args.model)

    # Model config
    model.names = set(list(MODEL_CONFIG_MAP[args.model_type].serving_aliases) + [args.model_type])
    model.max_length = MODEL_CONFIG_MAP[args.model_type].model_max_context
    model.eot_tokens = ray.get(tokenizer.get_eot_tokens.remote())

    model.enable_sys_prompt = args.enable_sys_prompt
    model.stream_period = args.stream_period
    model.api_keys = args.api_keys

    # Set max num batched tokens
    args.max_num_batched_tokens = max(args.max_num_batched_tokens or model.max_length, model.max_length)
    args.max_model_len = model.max_length

    # Load model engine
    engine_args = AsyncEngineArgs.from_cli_args(args)
    engine = AsyncLLMEngine.from_engine_args(engine_args)
    engine_model_config = asyncio.run(engine.get_model_config())

    # Run
    uvicorn.run(app,
                host=args.host,
                port=args.port,
                log_level="info",
                access_log=False,
                timeout_keep_alive=TIMEOUT_KEEP_ALIVE)



from typing import Optional, Callable, Iterable, List

from pydantic import BaseModel


class Message(BaseModel):
    role: str
    content: str

    weight: Optional[float] = None


class Conversation(BaseModel):
    items: List[Message]

    condition: str = ""
    system: str = ""


class ConversationTemplate(BaseModel):
    tokenizer: Callable

    # Prompt
    role_prefix: Callable
    system_as_role: Optional[bool] = False
    eot: str

    inference_condition: Optional[str] = None

    # Private
    bos_tokens_: List[int]
    eot_tokens_: List[int]

    def __init__(self, **data):
        tokenizer = data["tokenizer"]
        eot = data["eot"]
        bos_tokens_ = tokenizer("").input_ids
        eot_tokens_ = tokenizer(eot, add_special_tokens=False).input_ids
        super().__init__(**data, bos_tokens_=bos_tokens_, eot_tokens_=eot_tokens_)

    def _tokenize(self, strings: Iterable[str], ignore_special: bool = True) -> List[List[int]]:
        if self.tokenizer.is_fast:
            # Support for fast tokenizer
            # https://github.com/huggingface/tokenizers/pull/1419
            self.tokenizer._tokenizer.encode_special_tokens = ignore_special
            result = self.tokenizer(strings, return_attention_mask=False, add_special_tokens=False).input_ids
            self.tokenizer._tokenizer.encode_special_tokens = False
        else:
            result = self.tokenizer(strings, split_special_tokens=ignore_special, return_attention_mask=False, add_special_tokens=False).input_ids

        return result

    def tokenize_conversations(self, conversations: Iterable[Conversation], inference: bool = False, seq_level_weight: bool = False):
        # Pre-tokenize all conversations
        default_condition = self.inference_condition if inference else ""

        sys_mappings = set()
        role_mappings = set()
        all_text = []
        for conv in conversations:
            sys_mappings.add(conv.system)
            for msg in conv.items:
                role_mappings.add((msg.role, conv.condition or default_condition))
                all_text.append(msg.content)

        system_role_tokens = []
        if self.system_as_role:
            system_role_tokens = self._tokenize(self.role_prefix("system", ""), ignore_special=False)
        
        sys_mappings = list(sys_mappings)
        role_mappings = list(role_mappings)

        sys_mappings = dict(zip(sys_mappings, self._tokenize(sys_mappings)))
        role_mappings = dict(zip(role_mappings, self._tokenize([self.role_prefix(*args) for args in role_mappings], ignore_special=False)))
        all_text = self._tokenize(all_text)

        # Convert
        result_tokens = []
        result_weights = []
        all_text_idx = 0
        for conv in conversations:
            tokens = []
            weights = []

            # bos tokens
            tokens.extend(self.bos_tokens_)
            weights.extend([0.] * len(self.bos_tokens_))

            # System
            if conv.system:
                tokens.extend(system_role_tokens)
                weights.extend([0.] * len(system_role_tokens)) 
                
                system = sys_mappings[conv.system]
                tokens.extend(system)
                weights.extend([0.] * len(system))

                tokens.extend(self.eot_tokens_)
                weights.extend([0.] * len(self.eot_tokens_))

            # Messages
            last_idx = len(conv.items) - 1
            for idx, msg in enumerate(conv.items):
                # Role Prefix
                role = role_mappings[(msg.role, conv.condition or default_condition)]
                tokens.extend(role)
                weights.extend([0.] * len(role))
                
                # Message
                text = all_text[all_text_idx]
                all_text_idx += 1

                # weight
                w = None
                if not inference:
                    assert msg.weight is not None

                    w = msg.weight
                    if seq_level_weight:
                        w /= len(text) + len(self.eot_tokens_)

                # Message tokens
                tokens.extend(text)
                weights.extend([w] * len(text))

                if not (inference and idx == last_idx):  # Do not add EOT on last turn during inference
                    tokens.extend(self.eot_tokens_)
                    weights.extend([w] * len(self.eot_tokens_))

            # Append result
            result_tokens.append(tokens)
            result_weights.append(weights)

        # Sanity check
        assert all_text_idx == len(all_text)

        return result_tokens, result_weights



import argparse
import os
import math
import json
from functools import partial

import torch
import torch.distributed as dist
from torch.utils.data import DataLoader

import tqdm
import wandb
import numpy as np

from ochat.config import MODEL_CONFIG_MAP
from ochat.training_deepspeed.openchat_dataset import OpenchatDataset
from ochat.training_deepspeed.hf_hub import hub_upload_check, hub_upload_model_async

try:
    import deepspeed
except ImportError:
    raise ImportError("Please install deepspeed to train models.")


def parse_args():
    parser = argparse.ArgumentParser()
    # Distributed
    parser.add_argument("--local_rank", type=int, required=True)

    # Model type and data
    parser.add_argument("--model_path", type=str, required=True)
    parser.add_argument("--data_prefix", type=str, required=True)
    parser.add_argument("--save_path",  type=str, required=True)
    parser.add_argument("--save_every", type=int, default=None)
    parser.add_argument("--push_to_hub", type=str, default=None, 
                        help="Specify repository prefix for pushing to HuggingFace Hub. "
                             "For example, 'openchat/openchat-3.6' will create repositories "
                             "like 'openchat/openchat-3.6-ep0', 'openchat/openchat-3.6-ep1', ..."
                                "If not specified, will not push to Hub.")
    parser.add_argument("--push_to_hub_delete_local", action="store_true")

    # Hyperparameters
    parser.add_argument("--batch_max_len",      type=int, default=81920)
    parser.add_argument("--epochs",             type=int,   default=5)

    # Set lr to None to automatically estimate from LLaMA pretraining parameters (e.g. lr ~ sqrt(batch_size))
    parser.add_argument("--lr",                 type=float, default=None)
    parser.add_argument("--lr_min_ratio",       type=float, default=0.1)
    parser.add_argument("--lr_warmup_ratio",    type=int,   default=0.05)

    parser.add_argument("--weight_decay",       type=float, default=0.1)

    parser.add_argument("--beta1",              type=float, default=0.9)
    parser.add_argument("--beta2",              type=float, default=0.95)
    parser.add_argument("--eps",                type=float, default=1e-5)
    
    parser.add_argument("--wandb_entity",       type=str, default=None)
    parser.add_argument("--wandb_project",      type=str, default=None)

    # DeepSpeed parameters
    parser = deepspeed.add_config_arguments(parser)

    # Parse known args
    args, unknown = parser.parse_known_args()
    return args


def create_dataset_and_dataloader(args, epoch: int):
    # Find data
    filename = f"{args.data_prefix}.{epoch}.parquet"

    # Create dataset and dataloader
    print(f"Loading epoch {epoch} data from {filename}...")

    dataset = OpenchatDataset(
        dataset_filename=filename,

        batch_max_length=args.batch_max_len,
        rank=dist.get_rank(),
        num_replicas=dist.get_world_size()
    )
    dataloader = DataLoader(
        dataset,
        batch_size=None,

        num_workers=1,
        prefetch_factor=8,

        pin_memory=True
    )
    return dataset, dataloader


def create_model(args):
    print(f"Loading model {args.model_type} from {args.model_path}...")

    # Create model + optimizer + lr scheduler
    model = MODEL_CONFIG_MAP[args.model_type].model_create_for_training(args.model_path)
    # Model to assigned cuda device
    model = model.to(args.local_rank)
    # Enable gradient checkpointing
    model.gradient_checkpointing_enable(gradient_checkpointing_kwargs=dict(
        use_reentrant=False
    ))

    # Optimizer
    optimizer = torch.optim.AdamW(model.parameters(),
                                  lr=args.lr,
                                  weight_decay=args.weight_decay,
                                  betas=(args.beta1, args.beta2),
                                  eps=args.eps,
                                  fused=True)

    # DeepSpeed model
    model_engine, optimizer, _, _ = deepspeed.initialize(args=args,
                                                         model=model,
                                                         model_parameters=model.parameters(),
                                                         optimizer=optimizer)

    # Put deepspeed arguments
    args.device                         = model_engine.device

    return model_engine, optimizer


def cosine_schedule_with_warmup_lr_lambda(
    current_step: int, *, num_warmup_steps: int, num_training_steps: int, min_ratio: float = 0.0, num_cycles: float = 0.5
):
    if current_step < num_warmup_steps:
        return float(current_step) / float(max(1, num_warmup_steps))

    progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
    return min_ratio + max(0.0, (1 - min_ratio) * 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))


def create_lr_scheduler(args, train_total_steps):
    lr_scheduler = partial(
        cosine_schedule_with_warmup_lr_lambda,

        num_warmup_steps=round(args.lr_warmup_ratio * train_total_steps),
        num_training_steps=train_total_steps,
        min_ratio=args.lr_min_ratio
    )

    return lr_scheduler


def save_tokenizer(args, save_path):
    model_config = MODEL_CONFIG_MAP[args.model_type]
    tokenizer = model_config.model_tokenizer_create(args.model_path)
    tokenizer.chat_template = model_config.hf_chat_template
    tokenizer.save_pretrained(save_path)


def save_openchat_metadata(args, epoch, save_path):
    metadata = vars(args)
    metadata["epoch"] = epoch

    with open(os.path.join(save_path, "openchat.json"), "w") as f:
        json.dump(metadata, f, default=lambda o: "<non-serializable>")


def calculate_auto_lr(lr, batch_max_len, model_type, train_dataset):
    if lr is not None:
        return lr
    
    # Llama hyperparameters
    # FIXME: Only 7B/13B is supported
    base_lr = 3e-4
    base_bs = 4_000_000
    if "mistral" in model_type.lower():
        base_lr /= 6.0
    elif "gemma" in model_type.lower():
        base_lr /= 5.5  # NOTE(one): Maybe MLP and Attn layers are using different lr?
    elif "openchat_3.6" in model_type.lower():  # Llama 3 estimated hyperparams
        # NOTE(one): Estimated divisor: 1.5 * sqrt(25000 H100s / 2000 H100s)
        base_lr /= 5.3

    loss_weights = np.concatenate(train_dataset.dataset["nz_shifted_loss_weights"])
    supervised_ratio = np.sum(loss_weights != 0) / len(loss_weights)

    supervised_tokens = batch_max_len * dist.get_world_size() * supervised_ratio
    lr = base_lr * math.sqrt(supervised_tokens / base_bs)

    print(f"Use automatic learning rate {lr} (estimated from supervised ratio {supervised_ratio} effective batch size {supervised_tokens})")
    return lr


def state_dict_to_cpu(item, device=torch.device('cpu')):
    # Move all tensors to CPU
    if torch.is_tensor(item):
        return item.detach().to(device)
    elif isinstance(item, list):
        return [state_dict_to_cpu(v, device) for v in item]
    elif isinstance(item, tuple):
        return tuple([state_dict_to_cpu(v, device) for v in item])
    elif isinstance(item, dict):
        return type(item)({k: state_dict_to_cpu(v, device) for k, v in item.items()})
    else:
        return item


def train():
    deepspeed.init_distributed(dist_backend="nccl")
    RANK = dist.get_rank()
    WORLD_SIZE = dist.get_world_size()

    # Args
    args       = parse_args()

    hub_upload_check(args.push_to_hub)

    # Dataset
    train_dataset, train_loader = create_dataset_and_dataloader(args, 0)

    if train_dataset is None:
        raise RuntimeError("Training data not found.")

    # Load model type
    args.model_type = train_dataset.metadata["model_type"]

    train_total_steps = args.epochs * train_dataset.estimate_num_batches()

    # Hyperparams
    args.lr = calculate_auto_lr(args.lr, args.batch_max_len, args.model_type, train_dataset)

    # Model
    model_engine, optimizer = create_model(args)

    # LR Scheduler
    lr_scheduler = create_lr_scheduler(args, train_total_steps)

    # Progress bar and logger
    progress_bar = None
    if RANK == 0:
        progress_bar = tqdm.tqdm(total=train_total_steps)

        wandb.init(project=args.wandb_project or os.path.basename(args.model_path), entity=args.wandb_entity, config=args)

    # Training Loop
    step = 0
    lr_this_step = None
    for epoch in range(args.epochs):
        print (f"[rank {RANK} of {WORLD_SIZE}]: Epoch {epoch}")

        ############ Load Dataset
        if epoch != 0:
            del train_dataset, train_loader

            train_dataset, train_loader = create_dataset_and_dataloader(args, epoch)

        ############ Train Epoch
        model_engine.train()
        for (batch_tensor, batch_info), all_numseq, cur_numseq in train_loader:
            step += 1
            if step > train_total_steps:  # At most train_total_steps
                break

            # To device
            batch_tensor = {k: (v.to(args.device) if v is not None else None) for k, v in batch_tensor.items()}

            # Update
            loss, acc = model_engine(**batch_tensor, **batch_info).loss
            loss = (WORLD_SIZE / all_numseq) * loss
            acc  = (WORLD_SIZE / all_numseq) * acc

            model_engine.backward(loss)

            if model_engine.is_gradient_accumulation_boundary():
                # Set LR
                lr_this_step = args.lr * lr_scheduler(step)
                for param_group in optimizer.param_groups:
                    param_group['lr'] = lr_this_step

            model_engine.step()

            # Logging
            if RANK == 0:
                wandb.log({
                    "train/loss": loss.item() * (all_numseq / (cur_numseq * WORLD_SIZE)),
                    "train/acc":  acc.item()  * (all_numseq / (cur_numseq * WORLD_SIZE)),
                    "train/lr": lr_this_step
                }, step=step)
                progress_bar.update()  # type: ignore

        ############ Save Checkpoint
        # Save model with lean state dict
        # https://deepspeed.readthedocs.io/en/latest/model-checkpointing.html
        if (epoch + 1 == args.epochs) or (args.save_every and ((epoch + 1) % args.save_every == 0)):
            if RANK == 0:
                save_path = os.path.join(args.save_path, f"ep_{epoch}")

                model_engine.module.save_pretrained(save_path,
                                                    state_dict=state_dict_to_cpu(model_engine.module.state_dict()))  # type: ignore

                # Also save tokenizer from base model
                save_tokenizer(args, save_path)

                # Write metadata
                save_openchat_metadata(args, epoch, save_path)

                # Upload to hub
                hub_upload_model_async(
                    args.push_to_hub,
                    args.push_to_hub_delete_local,
                    save_path,
                    epoch
                )


if __name__ == "__main__":
    train()



# Adapted from
# https://github.com/lm-sys/FastChat/blob/168ccc29d3f7edc50823016105c024fe2282732a/fastchat/protocol/openai_api_protocol.py
import time
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from vllm.utils import random_uuid


class ErrorResponse(BaseModel):
    object: str = "error"
    message: str
    type: str
    param: Optional[str] = None
    code: Optional[str] = None


class ModelPermission(BaseModel):
    id: str = Field(default_factory=lambda: f"modelperm-{random_uuid()}")
    object: str = "model_permission"
    created: int = Field(default_factory=lambda: int(time.time()))
    allow_create_engine: bool = False
    allow_sampling: bool = True
    allow_logprobs: bool = False
    allow_search_indices: bool = False
    allow_view: bool = True
    allow_fine_tuning: bool = False
    organization: str = "*"
    group: Optional[str] = None
    is_blocking: str = False


class ModelCard(BaseModel):
    id: str
    object: str = "model"
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: str = "openchat"
    root: Optional[str] = None
    parent: Optional[str] = None
    permission: List[ModelPermission] = Field(default_factory=list)


class ModelList(BaseModel):
    object: str = "list"
    data: List[ModelCard] = Field(default_factory=list)


class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0
    completion_tokens: Optional[int] = 0


class ChatCompletionRequest(BaseModel):
    model: str
    messages: Union[str, List[Dict[str, str]]]
    condition: Optional[str] = ""
    temperature: Optional[float] = 0.7
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    max_tokens: Optional[int] = None
    seed: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    stream: Optional[bool] = False
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None


class ChatMessage(BaseModel):
    role: str
    content: str


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatMessage
    finish_reason: Optional[Literal["stop", "length"]] = None


class ChatCompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"chatcmpl-{random_uuid()}")
    object: str = "chat.completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo


class DeltaMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = None


class ChatCompletionResponseStreamChoice(BaseModel):
    index: int
    delta: DeltaMessage
    finish_reason: Optional[Literal["stop", "length"]] = None


class ChatCompletionStreamResponse(BaseModel):
    id: str = Field(default_factory=lambda: f"chatcmpl-{random_uuid()}")
    object: str = "chat.completion.chunk"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseStreamChoice]


class LoggingRecord(BaseModel):
    time: int
    request: ChatCompletionRequest
    outputs: List[str]



from typing import Optional, Callable, Iterable

from pydantic import BaseModel, ConfigDict


class ModelConfig(BaseModel):
    # Alias
    serving_aliases: Iterable[str] = ()

    # Model
    model_max_context: int
    model_tokenizer_create: Callable
    model_create_for_training: Callable

    # conversation template
    conversation_template: Callable
    hf_chat_template: Optional[str] = None

    model_config = ConfigDict(protected_namespaces=())  # Disables warnings for the model_ namespace used above



import argparse

import transformers
import torch


def modify_eos_embeddings(model_path, output_dir):
    tokenizer = transformers.AutoTokenizer.from_pretrained(model_path)
    model = transformers.AutoModelForCausalLM.from_pretrained(model_path, low_cpu_mem_usage=True, torch_dtype=torch.bfloat16)

    eos_token_id = tokenizer.eos_token_id

    print (f"EOS Token {tokenizer.convert_ids_to_tokens(eos_token_id)} ID {eos_token_id}")
    with torch.no_grad():
        model.model.embed_tokens.weight[eos_token_id] = torch.mean(model.model.embed_tokens.weight, dim=0)
        model.lm_head.weight[eos_token_id]            = torch.mean(model.lm_head.weight, dim=0)

    # Save
    tokenizer.save_pretrained(output_dir)
    model.save_pretrained(output_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-path",
        help="Location of model, or HuggingFace repo ID",
    )
    parser.add_argument(
        "--output-dir",
        help="Location to write resulting model and tokenizer",
    )

    modify_eos_embeddings(**vars(parser.parse_args()))


if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright 2022 EleutherAI and the HuggingFace Inc. team. All rights reserved.
#
# This code is based on EleutherAI's GPT-NeoX library and the GPT-NeoX
# and OPT implementations in this library. It has been modified from its
# original forms to accommodate minor architectural differences compared
# to GPT-NeoX and OPT used by the Meta AI team that trained the model.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" PyTorch Unpadded & Fused Gemma model. Compatible with HF. """

from typing import Optional, Tuple

import torch
import torch.utils.checkpoint
import torch.nn.functional as F
from torch import nn

from transformers.activations import ACT2FN
from transformers.modeling_outputs import CausalLMOutputWithPast
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import logging
from transformers.models.gemma.configuration_gemma import GemmaConfig

try:
    from flash_attn.flash_attn_interface import flash_attn_varlen_func
    from flash_attn.bert_padding import pad_input
except ImportError:
    print ("FlashAttention not found. Install it if you need to train models.")


def rotate_half(x: torch.Tensor):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_pos_emb(q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor, position_ids: torch.Tensor):
    # q, k:     [nnz, num_heads, head_dim]
    # position_ids: [nnz]
    # cos, sin: [max_seq_len, head_dim]
    cos = cos[position_ids].unsqueeze(-2)  # [nnz, 1, head_dim]
    sin = sin[position_ids].unsqueeze(-2)  # [nnz, 1, head_dim]
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed


@torch.jit.script
def lm_head_with_loss(embed_weights: torch.Tensor, hidden_states: torch.Tensor, nz_shifted_label_ids: torch.Tensor, nz_shifted_loss_weights: torch.Tensor):
    logits = F.linear(hidden_states, embed_weights)

    loss = (nz_shifted_loss_weights * torch.nn.functional.cross_entropy(logits, nz_shifted_label_ids, reduction="none")).sum()
    token_accuracy = (nz_shifted_loss_weights * (torch.argmax(logits.detach(), dim=-1) == nz_shifted_label_ids)).sum()
    return loss, token_accuracy


# Copied from transformers.models.llama.modeling_llama.LlamaRMSNorm with Llama->Gemma
RMS_NORM_TRACED = None


def rms_norm(hidden_states: torch.Tensor, weight: torch.Tensor, variance_epsilon: torch.Tensor):
    input_dtype = hidden_states.dtype
    hidden_states = hidden_states.to(torch.float32)

    variance = hidden_states.square().mean(-1, keepdim=True)
    hidden_states = hidden_states * torch.rsqrt(variance + variance_epsilon)
    return (1 + weight) * hidden_states.to(input_dtype)


class UnpaddedGemmaRMSNorm(nn.Module):
    def __init__(self, hidden_size, eps):
        """
        UnpaddedGemmaRMSNorm is equivalent to T5LayerNorm
        """
        super().__init__()

        self.weight = nn.Parameter(torch.zeros(hidden_size))
        self.variance_epsilon = torch.tensor(eps, dtype=torch.get_default_dtype())

        global RMS_NORM_TRACED
        if RMS_NORM_TRACED is None:
            RMS_NORM_TRACED = torch.jit.trace(rms_norm, (torch.ones(hidden_size), torch.ones(hidden_size), self.variance_epsilon))

    def forward(self, hidden_states):
        global RMS_NORM_TRACED
        return RMS_NORM_TRACED(hidden_states, self.weight, self.variance_epsilon)


# Copied from transformers.models.llama.modeling_llama.LlamaRotaryEmbedding with Llama->Gemma
class UnpaddedGemmaRotaryEmbedding(torch.nn.Module):
    def __init__(self, dim, max_position_embeddings, base, device=None):
        super().__init__()

        # RoPE
        inv_freq = 1.0 / (base ** (torch.arange(0, dim, 2, dtype=torch.float32, device=device) / dim))
        t = torch.arange(max_position_embeddings, dtype=torch.float32, device=device)
        freqs = torch.outer(t, inv_freq)

        # Different from paper, but it uses a different permutation in order to obtain the same calculation
        emb = torch.cat((freqs, freqs), dim=-1)
        dtype = torch.get_default_dtype()
        self.register_buffer("cos_cached", emb.cos().to(dtype), persistent=False)
        self.register_buffer("sin_cached", emb.sin().to(dtype), persistent=False)

    def forward(self):
        return self.cos_cached, self.sin_cached


class UnpaddedGemmaMLP(nn.Module):
    def __init__(self, config: GemmaConfig):
        super().__init__()

        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size

        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=False)
        self.act_fn = ACT2FN[config.hidden_act]

    def forward(self, x):
        return self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x))


class UnpaddedGemmaAttention(nn.Module):
    """Multi-headed attention from 'Attention Is All You Need' paper"""

    def __init__(self, config: GemmaConfig):
        super().__init__()

        self.hidden_size = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.head_dim = config.head_dim
        self.num_key_value_heads = config.num_key_value_heads

        if self.hidden_size % self.num_heads != 0:
            raise ValueError(
                f"hidden_size must be divisible by num_heads (got `hidden_size`: {self.hidden_size}"
                f" and `num_heads`: {self.num_heads})."
            )

        self.q_proj = nn.Linear(self.hidden_size, self.num_heads * self.head_dim, bias=config.attention_bias)
        self.k_proj = nn.Linear(self.hidden_size, self.num_key_value_heads * self.head_dim, bias=config.attention_bias)
        self.v_proj = nn.Linear(self.hidden_size, self.num_key_value_heads * self.head_dim, bias=config.attention_bias)
        self.o_proj = nn.Linear(self.num_heads * self.head_dim, self.hidden_size, bias=config.attention_bias)

    def forward(
        self,
        cos_sin: Tuple[torch.Tensor, torch.Tensor],
        # Unpadded inputs
        nz_hidden_states: torch.Tensor,
        nz_position_ids: torch.LongTensor,
        cu_seqlens: torch.Tensor,
        max_seqlen: int
    ) -> torch.Tensor:
        # nz_hidden_states: [nnz, num_heads, head_dim]
        # nz_position_ids:  [nnz]
        # cu_seqlens:       [bs + 1]

        query_states = self.q_proj(nz_hidden_states).view(-1, self.num_heads, self.head_dim)
        key_states = self.k_proj(nz_hidden_states).view(-1,   self.num_key_value_heads, self.head_dim)
        value_states = self.v_proj(nz_hidden_states).view(-1, self.num_key_value_heads, self.head_dim)

        # RoPE
        cos, sin = cos_sin
        query_states, key_states = apply_rotary_pos_emb(query_states, key_states, cos, sin, nz_position_ids)

        # flash attn
        attn_output = flash_attn_varlen_func(
            q=query_states, k=key_states, v=value_states,
            cu_seqlens_q=cu_seqlens, cu_seqlens_k=cu_seqlens,
            max_seqlen_q=max_seqlen, max_seqlen_k=max_seqlen,

            dropout_p=0.0, causal=True)

        # attn_output: [total_nnz, num_heads, head_dim]
        attn_output = attn_output.view(-1, self.num_heads * self.head_dim)  # type: ignore
        return self.o_proj(attn_output)


class UnpaddedGemmaDecoderLayer(nn.Module):
    def __init__(self, config: GemmaConfig):
        super().__init__()

        self.hidden_size = config.hidden_size
        self.self_attn = UnpaddedGemmaAttention(config=config)
        self.mlp = UnpaddedGemmaMLP(config=config)
        self.input_layernorm = UnpaddedGemmaRMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = UnpaddedGemmaRMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def forward(
        self,
        cos_sin: Tuple[torch.Tensor, torch.Tensor],
        # Unpadded inputs
        nz_hidden_states: torch.Tensor,
        nz_position_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        max_seqlen: int
    ) -> torch.Tensor:
        # Self Attention
        residual = nz_hidden_states

        nz_hidden_states = self.input_layernorm(nz_hidden_states)
        nz_hidden_states = self.self_attn(
            cos_sin=cos_sin,

            nz_hidden_states=nz_hidden_states,
            nz_position_ids=nz_position_ids,
            cu_seqlens=cu_seqlens,
            max_seqlen=max_seqlen
        )
        nz_hidden_states = residual + nz_hidden_states

        # Fully Connected
        residual = nz_hidden_states

        nz_hidden_states = self.post_attention_layernorm(nz_hidden_states)
        nz_hidden_states = self.mlp(nz_hidden_states)
        nz_hidden_states = residual + nz_hidden_states

        return nz_hidden_states


class UnpaddedGemmaPreTrainedModel(PreTrainedModel):
    config_class = GemmaConfig
    base_model_prefix = "model"
    supports_gradient_checkpointing = True
    _no_split_modules = ["UnpaddedGemmaDecoderLayer"]

    def _init_weights(self, module):
        std = self.config.initializer_range
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()


class UnpaddedGemmaModel(UnpaddedGemmaPreTrainedModel):
    """
    Transformer decoder consisting of *config.num_hidden_layers* layers. Each layer is a [`UnpaddedGemmaDecoderLayer`]

    Args:
        config: GemmaConfig
    """

    def __init__(self, config: GemmaConfig):
        super().__init__(config)
        self.padding_idx = config.pad_token_id
        self.vocab_size = config.vocab_size
        self.normalization_factor = config.hidden_size ** 0.5

        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size, self.padding_idx)
        self.rotary_emb   = UnpaddedGemmaRotaryEmbedding(config.head_dim,
                                                         max_position_embeddings=config.max_position_embeddings,
                                                         base=config.rope_theta)

        self.layers = nn.ModuleList([UnpaddedGemmaDecoderLayer(config) for _ in range(config.num_hidden_layers)])
        self.norm = UnpaddedGemmaRMSNorm(config.hidden_size, eps=config.rms_norm_eps)

        self.gradient_checkpointing = False
        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.embed_tokens

    def set_input_embeddings(self, value):
        self.embed_tokens = value

    def forward(
        self,
        # Unpadded inputs
        nz_input_ids: torch.Tensor,
        nz_position_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        max_seqlen: int,
    ) -> torch.Tensor:
        nz_hidden_states = self.embed_tokens(nz_input_ids) * self.normalization_factor  # Normalized
        cos_sin          = self.rotary_emb()

        # decoder layers
        for decoder_layer in self.layers:
            if self.gradient_checkpointing and self.training:
                nz_hidden_states = self._gradient_checkpointing_func(
                    decoder_layer.__call__,

                    cos_sin,
                    nz_hidden_states,
                    nz_position_ids,
                    cu_seqlens,
                    max_seqlen
                )
            else:
                nz_hidden_states = decoder_layer(
                    cos_sin,
                    
                    nz_hidden_states,
                    nz_position_ids,
                    cu_seqlens,
                    max_seqlen
                )

        nz_hidden_states = self.norm(nz_hidden_states)

        return nz_hidden_states


class GemmaForCausalLM(UnpaddedGemmaPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.model = UnpaddedGemmaModel(config)

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.model.embed_tokens

    def set_input_embeddings(self, value):
        self.model.embed_tokens = value

    def get_output_embeddings(self):
        return self.model.embed_tokens

    def set_output_embeddings(self, new_embeddings):
        self.model.embed_tokens = new_embeddings

    def set_decoder(self, decoder):
        self.model = decoder

    def get_decoder(self):
        return self.model
    
    def forward(
        self,
        # Unpadded inputs
        nz_input_ids: torch.Tensor,
        nz_position_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        max_seqlen: int,
        # Unpadded labels
        nz_shifted_label_ids: Optional[torch.Tensor] = None,
        nz_shifted_loss_weights:      Optional[torch.Tensor] = None
    ) -> CausalLMOutputWithPast:
        # Model logits
        hidden_states = self.model(
            nz_input_ids=nz_input_ids,
            nz_position_ids=nz_position_ids,
            cu_seqlens=cu_seqlens,
            max_seqlen=max_seqlen
        )

        # Loss
        loss = lm_head_with_loss(
            self.model.embed_tokens.weight,  # Tied embeddings
            hidden_states,
            nz_shifted_label_ids,
            nz_shifted_loss_weights
        )

        return CausalLMOutputWithPast(
            loss=loss  # type: ignore
        )



import ray

from ochat.config import Message, Conversation


@ray.remote
class AsyncTokenizer:
    def __init__(self, model_type: str, model_path: str) -> None:
        from ochat.config import MODEL_CONFIG_MAP

        config = MODEL_CONFIG_MAP[model_type]
        tokenizer = config.model_tokenizer_create(model_path)

        self.conv_template = config.conversation_template(tokenizer=tokenizer)

    def tokenize(self, messages, condition, enable_sys_prompt=False):
        # get system messages
        system_message = ""
        items = []

        for msg_raw in messages:
            msg = Message(**msg_raw)
            if msg.role == "system":
                # Use system prompt only when enabled
                if enable_sys_prompt:
                    system_message = msg.content.strip()

                continue

            items.append(msg)

        assert len(items)

        # append ai role
        if items[-1].role != "assistant":
            items.append(Message(role="assistant", content=""))

        tokens, _ = self.conv_template.tokenize_conversations([Conversation(items=items, system=system_message, condition=condition)],
                                                              inference=True)
        return tokens[0]
    
    def get_eot_tokens(self):
        assert len(self.conv_template.eot_tokens_) == 1

        return self.conv_template.eot_tokens_



"""
Generate training data based on conversations

Usage: python -m ochat.data.generate_data --in-file sharegpt_gpt4.jsonl --tokenizer-name HF_REPO_NAME --out-dir .
"""

import argparse
import os
import gc
import random

import ray
import orjson
import pyarrow
from pyarrow import parquet


PAD_TOKEN_ID = 0


def _split(a, n):
    # Split list a to n chunks
    # https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length
    k, m = divmod(len(a), n)
    return [a[i*k+min(i, m): (i+1)*k+min(i+1, m)] for i in range(n)]


def truncate_trailing_zero_weighted(tokens, weights):
    non_zero_index = len(weights) - 1
    while non_zero_index >= 0 and weights[non_zero_index] == 0:
        non_zero_index -= 1

    return tokens[:non_zero_index + 1], weights[:non_zero_index + 1]


def add_single_conv(output, tokens, weights):
    # truncate trailing zero weighted tokens
    tokens, weights = truncate_trailing_zero_weighted(tokens, weights)
    if not tokens:
        return

    # labels
    length = len(tokens)
    labels = [(t if w != 0 else PAD_TOKEN_ID) for t, w in zip(tokens, weights)]

    # populate results
    results = {
        "total_length": length,

        "seqlens": [length],
        "nz_input_ids": tokens,
        "nz_position_ids": list(range(length)),

        "nz_shifted_label_ids":    labels[1:]  + [PAD_TOKEN_ID],
        "nz_shifted_loss_weights": weights[1:] + [0.0]
    }
    results["num_seqs"] = sum(results["nz_shifted_loss_weights"])

    for k, v in results.items():
        output[k].append(v)


@ray.remote
def convert_conversation_batch(model_type: str, model_path: str, batch: list, schema: pyarrow.Schema, per_sequence_loss: bool):
    from ochat.config import MODEL_CONFIG_MAP, Conversation

    # Tokenization
    model_config = MODEL_CONFIG_MAP[model_type]
    tokenizer = model_config.model_tokenizer_create(model_path)
    conv_template = model_config.conversation_template(tokenizer=tokenizer)

    # Decode data
    print ("Decoding JSON ...")
    batch = [Conversation(**orjson.loads(json_line)) for json_line in batch]

    # Tokenize
    print ("Tokenizing ...")
    tokens_list, weights_list = conv_template.tokenize_conversations(batch, inference=False, seq_level_weight=per_sequence_loss)

    del batch
    gc.collect()

    # Generate data
    print ("Generating ...")
    max_context = model_config.model_max_context

    outputs = {k: [] for k in schema.names}
    for tokens, weights in zip(tokens_list, weights_list):
        assert len(tokens) == len(weights)

        # Truncate to specified tokens
        tokens  = tokens[:max_context]
        weights = weights[:max_context]

        # Add to results
        add_single_conv(outputs, tokens, weights)

    del tokens_list, weights_list
    gc.collect()

    print ("To table ...")
    table = pyarrow.Table.from_pydict(outputs, schema=schema)

    del outputs
    gc.collect()

    print ("Chunk finish")
    return table


def generate_epoch(seed: int, model_type: str, model_path: str, in_filename: str, out_filename: str, per_sequence_loss: bool):
    # schema
    metadata = {
        "model_type": model_type
    }
    schema = [
        pyarrow.field("total_length", pyarrow.int32()),
        pyarrow.field("num_seqs", pyarrow.float32()),

        pyarrow.field(f"seqlens", pyarrow.list_(pyarrow.int32())),
        pyarrow.field(f"nz_input_ids", pyarrow.list_(pyarrow.int32())),
        pyarrow.field(f"nz_position_ids", pyarrow.list_(pyarrow.int32())),
        pyarrow.field(f"nz_shifted_label_ids", pyarrow.list_(pyarrow.int32())),
        pyarrow.field(f"nz_shifted_loss_weights", pyarrow.list_(pyarrow.float32()))
    ]

    schema = pyarrow.schema(schema, metadata={"metadata_json": orjson.dumps(metadata)})

    # Load data
    with open(in_filename, "rb") as f:
        batches = f.readlines()

        random.seed(seed)  # Randomized load balancing
        random.shuffle(batches)

        batches = _split(batches, int(ray.available_resources()["CPU"]))

    # launch remote workers
    handles = [convert_conversation_batch.remote(
        model_type=model_type,  # type: ignore
        model_path=model_path,
        batch=batch,
        schema=schema,
        per_sequence_loss=per_sequence_loss
    ) for batch in batches]

    # write
    parquet.write_table(pyarrow.concat_tables([ray.get(handle) for handle in handles]), out_filename)


def generate_dataset(model_type, model_path, in_prefix, out_prefix, per_sequence_loss, seed):
    # Initialize Ray
    if not ray.is_initialized():
        ray.init(ignore_reinit_error=True, num_cpus=os.cpu_count())

    # Load epochs and tokenize
    epoch = 0
    while True:
        in_filename = f"{in_prefix}.{epoch}.jsonl"
        if not os.path.exists(in_filename):
            break

        out_filename = f"{out_prefix}.{epoch}.parquet"
        generate_epoch(
            seed=seed + epoch,
            model_type=model_type,
            model_path=model_path,
            in_filename=in_filename,
            out_filename=out_filename,
            per_sequence_loss=per_sequence_loss
        )
        gc.collect()

        epoch += 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--model-type", type=str, required=True)
    parser.add_argument("--model-path", type=str, required=True)

    parser.add_argument("--in-prefix", type=str, required=True)
    parser.add_argument("--out-prefix", type=str, required=True)

    parser.add_argument("--per-sequence-loss", action="store_true")
    parser.add_argument("--seed", type=int, default=42)
    args = parser.parse_args()

    generate_dataset(**vars(args))



import argparse

import transformers
import torch


def add_tokens_to_embedding(added_special_tokens, embedding):
    # Mean embedding, shape: [1, dim]
    new_token_embeddings = torch.mean(embedding.to(torch.float32), dim=0, keepdim=True).to(embedding.dtype)
    # Expand to [N, dim]
    new_token_embeddings = new_token_embeddings.expand(len(added_special_tokens), -1)

    return torch.cat([embedding, new_token_embeddings], dim=0)


def hf_add_tokens(model_path, output_dir, added_special_tokens):
    tokenizer = transformers.AutoTokenizer.from_pretrained(model_path)
    model = transformers.AutoModelForCausalLM.from_pretrained(model_path,
                                                              low_cpu_mem_usage=True,
                                                              torch_dtype=torch.bfloat16)
    # Add tokens (tokenizer)
    tokenizer.add_special_tokens({"additional_special_tokens": added_special_tokens})

    # Add tokens (embedding)
    assert model.model.embed_tokens.weight.requires_grad
    assert model.lm_head.weight.requires_grad

    model.model.embed_tokens.weight = torch.nn.Parameter(add_tokens_to_embedding(added_special_tokens, model.model.embed_tokens.weight), requires_grad=True)
    model.lm_head.weight            = torch.nn.Parameter(add_tokens_to_embedding(added_special_tokens, model.lm_head.weight), requires_grad=True)

    model.config.vocab_size += len(added_special_tokens)

    # Fix model config (Mistral's actual token length is 8192)
    if "mistral" in model_path.lower():
        assert model.config.max_position_embeddings == 32768
        model.config.max_position_embeddings = 8192

    print ({k: v.shape for k, v in model.state_dict().items()})

    # Save
    tokenizer.save_pretrained(output_dir)
    model.save_pretrained(output_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-path",
        help="Location of Mistral model, or HuggingFace repo ID",
    )
    parser.add_argument(
        "--output-dir",
        help="Location to write resulting model and tokenizer",
    )
    parser.add_argument(
        "--added-special-tokens",
        type=str,
        nargs="+",
        help="Special token list to add"
    )

    hf_add_tokens(**vars(parser.parse_args()))


if __name__ == "__main__":
    main()



from typing import OrderedDict
import signal
import os
import json
import subprocess
import argparse
import time
import requests
import re
import coolname


MAX_CONTEXT = 4096


def find_models(path, prefix, ep_filter):
    run_name = '_'.join(coolname.generate(2))

    def generate_model_name(root, ep_number):
        return f"{prefix}{os.path.basename(root)}_ep{ep_number}_{run_name}"

    models = {}
    for root, dirs, _ in os.walk(path):
        for d in dirs:
            ep_match = re.match(r"ep_(\d+)", d)
            if not ep_match:
                continue

            if ep_filter and ep_match.group(1) != ep_filter:
                continue

            model_name = generate_model_name(root, ep_match.group(1))
            models[model_name] = os.path.join(root, d)

    # Sort and return the models dictionary as an OrderedDict
    return OrderedDict(sorted(models.items(), reverse=True, key=lambda x: x[0].split("_ep")[::-1]))


def run_mt_bench(mt_bench_path, model_name):
    working_dir = os.path.join(mt_bench_path, "fastchat", "llm_judge")

    # Skip if result exists
    if os.path.exists(os.path.join(working_dir, "data", "mt_bench", "model_answer", f"{model_name}.jsonl")):
        return

    # run mt bench
    commands = [
        f"python gen_api_answer.py --model {model_name} --max-tokens {MAX_CONTEXT} --parallel 128 --openai-api-base http://localhost:18888/v1",
        f"python gen_judgment.py --model-list {model_name} --parallel 8 --mode single",
        # f"python gen_judgment.py --model-list {model_name} --parallel 8 --mode pairwise-baseline",
    ]

    env = os.environ.copy()
    env["PYTHONPATH"] = f'{env.get("PYTHONPATH", "")}:{mt_bench_path}'

    for command in commands:
        subprocess.run(command, shell=True, cwd=working_dir, env=env)


def run_vicuna_bench(mt_bench_path, model_name):
    working_dir = os.path.join(mt_bench_path, "fastchat", "llm_judge")

    # Skip if result exists
    if os.path.exists(os.path.join(working_dir, "data", "vicuna_bench", "model_answer", f"{model_name}.jsonl")):
        return

    # run mt bench
    commands = [
        f"python gen_api_answer.py --model {model_name} --max-tokens {MAX_CONTEXT} --parallel 128 --openai-api-base http://localhost:18888/v1 --bench-name vicuna_bench",
        f"python gen_judgment.py --model-list {model_name} --parallel 8 --mode pairwise-baseline --bench-name vicuna_bench",
    ]

    env = os.environ.copy()
    env["PYTHONPATH"] = f'{env.get("PYTHONPATH", "")}:{mt_bench_path}'

    for command in commands:
        subprocess.run(command, shell=True, cwd=working_dir, env=env)


def create_alpaca_eval_config(alpacaeval_path, model_name):
    config_dir = os.path.join(alpacaeval_path, "src", "alpaca_eval", "models_configs", model_name.lower())
    os.makedirs(config_dir, exist_ok=True)
    config_path = os.path.join(config_dir, "configs.yaml")

    config_content = f"""{model_name.lower()}:
  prompt_template: "openchat-13b/prompt.txt"
  fn_completions: "openai_completions"
  completions_kwargs:
    openai_api_base: http://127.0.0.1:18888/v1
    requires_chatml: True
    sleep_time: 0

    model_name: "{model_name.lower()}"
    max_tokens: {MAX_CONTEXT}

    top_p: 1.0
    temperature: 0.7

    num_procs: 128

  pretty_name: "{model_name}"
  link: "https://github.com/imoneoi/openchat"
"""

    with open(config_path, "w") as f:
        f.write(config_content)


def run_alpaca_eval(alpacaeval_path, model_name):
    # Skip if result exists
    if os.path.exists(os.path.join(alpacaeval_path, "results", model_name.lower())):
        return

    # Create config
    create_alpaca_eval_config(alpacaeval_path, model_name)

    # Run
    command = f"python -m alpaca_eval.main evaluate_from_model --model_configs {model_name.lower()} --annotators_config alpaca_eval_gpt4"

    env = os.environ.copy()
    env["PYTHONPATH"] = f'{env.get("PYTHONPATH", "")}:{os.path.join(alpacaeval_path, "src")}'

    subprocess.run(command, shell=True, cwd=alpacaeval_path, env=env)


def wait_for_server(url):
    while True:
        try:
            response = requests.get(url)
            if response.status_code in [200, 404]:
                break
        except requests.exceptions.RequestException:
            pass

        time.sleep(1)


def main(path, prefix, ep_filter, mt_bench_path, alpacaeval_path):
    models = find_models(path, prefix, ep_filter)

    for i, (model_name, model_path) in enumerate(models.items()):
        print(f"Processing model {i + 1}/{len(models)}: {model_name}")

        print("Starting server...")
        server_command = f"python -m ochat.serving.openai_api_server --model {model_path} --engine-use-ray --worker-use-ray"
        server_process = subprocess.Popen(server_command, shell=True, preexec_fn=os.setsid)

        wait_for_server("http://127.0.0.1:18888/v1")
        print("Server is ready.")

        print("Running MT-bench...")
        run_mt_bench(mt_bench_path, model_name)

        print("Running AlpacaEval...")
        # run_alpaca_eval(alpacaeval_path, model_name)

        # print("Running Vicuna-bench")
        # run_vicuna_bench(mt_bench_path, model_name)

        print("Terminating server...")
        os.killpg(os.getpgid(server_process.pid), signal.SIGTERM)
        server_process.wait()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", default="/ML-A100/home/csj/trained_models/openchat_mistral/1017", help="Path to the models directory")
    parser.add_argument("--prefix", default="gpt4correct_")
    parser.add_argument("--ep_filter", default=None, help="Filter epochs")

    parser.add_argument("--mt_bench_path", default="/ML-A100/home/csj/one_benchmarks/FastChat", help="Path to the MT-bench directory")
    parser.add_argument("--alpacaeval_path", default="/ML-A100/home/csj/one_benchmarks/alpaca_eval", help="Path to the AlpacaEval directory")
    args = parser.parse_args()

    main(**vars(args))



import shlex
import subprocess

from huggingface_hub import HfApi


def hub_upload_check(push_to_hub: str):
    if push_to_hub is not None:
        # Try creating a test repo
        test_repo_name = f"{push_to_hub}-dummy-test"
        try:
            HfApi().create_repo(
                repo_id=test_repo_name,
                repo_type="model",
                private=True
            )
            HfApi().delete_repo(
                repo_id=test_repo_name,
                repo_type="model"
            )
        except Exception as e:
            raise RuntimeError(f"Failed to push test repo {test_repo_name} to HuggingFace Hub. Please check your permissions and network connection. Use `huggingface-cli login` to log in your account.\n\n{str(e)}")


def hub_upload_model_async(push_to_hub: str, push_to_hub_delete_local: bool, save_path: str, epoch: int):
    if push_to_hub is not None:
        safe_repo_name = shlex.quote(f"{push_to_hub}-ep-{epoch}")
        safe_save_path = shlex.quote(save_path)

        command = f"huggingface-cli upload --quiet --repo-type model --private {safe_repo_name} {safe_save_path}"
        if push_to_hub_delete_local:
            command += f" && rm -rf {safe_save_path}"

        subprocess.Popen(command, shell=True)



import argparse
import os
from pathlib import Path

import orjson
import pandas as pd
from glob import glob

def save_results(dfs, save_path: str):
    os.makedirs(os.path.dirname(save_path), exist_ok=True)
    with pd.ExcelWriter(save_path) as writer:
        for task_type, df_task in dfs.items():
            df_task.to_excel(writer, sheet_name=task_type)
                
def view_results(result_path: str):
    # Read results
    eval_results = []
    for filename in glob(os.path.join(result_path, "*.json")):
        with open(filename, "rb") as f:
            questions = orjson.loads(f.read())

            eval_results.extend([{
                "model": Path(filename).stem,
                "task_type": q["task_type"],
                "task_name": os.path.relpath(q["task_name"], q["task_type"]),
                "accuracy": q["is_correct"],
                "unmatched": not q["is_matched"],
            } for q in questions])
    df = pd.DataFrame.from_records(eval_results)
    all_tables = dict()    
    # Overall metrics table
    df_overall = df.pivot_table(index=["model"], columns=["task_type"], values=["accuracy", "unmatched"], aggfunc="mean")
    all_tables["overall"] = df_overall
    print(df_overall.to_string(float_format=lambda x: f"{x * 100:.1f}", na_rep="-"))
    # Print tables for each task
    for task_type in df["task_type"].unique():
        df_task = df[df["task_type"] == task_type].pivot_table(index=["task_name"], columns=["model"], values=["accuracy", "unmatched"], aggfunc="mean")
        all_tables[task_type.replace("/", "_")] = df_task
        print(f"\n### {task_type}\n")
        print(df_task.to_string(float_format=lambda x: f"{x * 100:.1f}", na_rep="-"))
    return all_tables


def main():
    parser = argparse.ArgumentParser()

    # Input / output
    parser.add_argument("--result_path", type=str, default="ochat/evaluation/eval_results")
    parser.add_argument("--save_path", type=str, default="ochat/evaluation/eval_results/summary.xlsx")
    parser.add_argument("--save", "-s", action="store_true", help="Save the results to a file")
    args = parser.parse_args()
    
    all_tables = view_results(args.result_path)
    if args.save:
        save_results(all_tables, args.save_path)


if __name__ == "__main__":
    main()



#    Copyright 2023 Rohan Taori, Ishaan Gulrajani, Tianyi Zhang, Yann Dubois, Xuechen Li
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import logging
from dataclasses import dataclass, field
from typing import Dict, Optional, Sequence

import torch
import transformers
import utils
from torch.utils.data import Dataset
from transformers import Trainer

IGNORE_INDEX = -100
DEFAULT_PAD_TOKEN = "[PAD]"
DEFAULT_EOS_TOKEN = "</s>"
DEFAULT_BOS_TOKEN = "<s>"
DEFAULT_UNK_TOKEN = "<unk>"
PROMPT_DICT = {
    "prompt_input": (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        "### Instruction:\n{instruction}\n\n### Input:\n{input}\n\n### Response:"
    ),
    "prompt_no_input": (
        "Below is an instruction that describes a task. "
        "Write a response that appropriately completes the request.\n\n"
        "### Instruction:\n{instruction}\n\n### Response:"
    ),
}


@dataclass
class ModelArguments:
    model_name_or_path: Optional[str] = field(default="facebook/opt-125m")


@dataclass
class DataArguments:
    data_path: str = field(default=None, metadata={"help": "Path to the training data."})


@dataclass
class TrainingArguments(transformers.TrainingArguments):
    cache_dir: Optional[str] = field(default=None)
    optim: str = field(default="adamw_torch")
    model_max_length: int = field(
        default=512,
        metadata={"help": "Maximum sequence length. Sequences will be right padded (and possibly truncated)."},
    )


def smart_tokenizer_and_embedding_resize(
    special_tokens_dict: Dict,
    tokenizer: transformers.PreTrainedTokenizer,
    model: transformers.PreTrainedModel,
):
    """Resize tokenizer and embedding.

    Note: This is the unoptimized version that may make your embedding size not be divisible by 64.
    """
    num_new_tokens = tokenizer.add_special_tokens(special_tokens_dict)
    model.resize_token_embeddings(len(tokenizer))

    if num_new_tokens > 0:
        input_embeddings = model.get_input_embeddings().weight.data
        output_embeddings = model.get_output_embeddings().weight.data

        input_embeddings_avg = input_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)
        output_embeddings_avg = output_embeddings[:-num_new_tokens].mean(dim=0, keepdim=True)

        input_embeddings[-num_new_tokens:] = input_embeddings_avg
        output_embeddings[-num_new_tokens:] = output_embeddings_avg


def _tokenize_fn(strings: Sequence[str], tokenizer: transformers.PreTrainedTokenizer) -> Dict:
    """Tokenize a list of strings."""
    tokenized_list = [
        tokenizer(
            text,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_length,
            truncation=True,
        )
        for text in strings
    ]
    input_ids = labels = [tokenized.input_ids[0] for tokenized in tokenized_list]
    input_ids_lens = labels_lens = [
        tokenized.input_ids.ne(tokenizer.pad_token_id).sum().item() for tokenized in tokenized_list
    ]
    return dict(
        input_ids=input_ids,
        labels=labels,
        input_ids_lens=input_ids_lens,
        labels_lens=labels_lens,
    )


def preprocess(
    sources: Sequence[str],
    targets: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
) -> Dict:
    """Preprocess the data by tokenizing."""
    examples = [s + t for s, t in zip(sources, targets)]
    examples_tokenized, sources_tokenized = [_tokenize_fn(strings, tokenizer) for strings in (examples, sources)]
    input_ids = examples_tokenized["input_ids"]
    labels = copy.deepcopy(input_ids)
    for label, source_len in zip(labels, sources_tokenized["input_ids_lens"]):
        label[:source_len] = IGNORE_INDEX
    return dict(input_ids=input_ids, labels=labels)


class SupervisedDataset(Dataset):
    """Dataset for supervised fine-tuning."""

    def __init__(self, data_path: str, tokenizer: transformers.PreTrainedTokenizer):
        super(SupervisedDataset, self).__init__()
        logging.warning("Loading data...")
        list_data_dict = utils.jload(data_path)

        logging.warning("Formatting inputs...")
        prompt_input, prompt_no_input = PROMPT_DICT["prompt_input"], PROMPT_DICT["prompt_no_input"]
        sources = [
            prompt_input.format_map(example) if example.get("input", "") != "" else prompt_no_input.format_map(example)
            for example in list_data_dict
        ]
        targets = [f"{example['output']}{tokenizer.eos_token}" for example in list_data_dict]

        logging.warning("Tokenizing inputs... This may take some time...")
        data_dict = preprocess(sources, targets, tokenizer)

        self.input_ids = data_dict["input_ids"]
        self.labels = data_dict["labels"]

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, i) -> Dict[str, torch.Tensor]:
        return dict(input_ids=self.input_ids[i], labels=self.labels[i])


@dataclass
class DataCollatorForSupervisedDataset(object):
    """Collate examples for supervised fine-tuning."""

    tokenizer: transformers.PreTrainedTokenizer

    def __call__(self, instances: Sequence[Dict]) -> Dict[str, torch.Tensor]:
        input_ids, labels = tuple([instance[key] for instance in instances] for key in ("input_ids", "labels"))
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids, batch_first=True, padding_value=self.tokenizer.pad_token_id
        )
        labels = torch.nn.utils.rnn.pad_sequence(labels, batch_first=True, padding_value=IGNORE_INDEX)
        return dict(
            input_ids=input_ids,
            labels=labels,
            attention_mask=input_ids.ne(self.tokenizer.pad_token_id),
        )


def make_supervised_data_module(tokenizer: transformers.PreTrainedTokenizer, data_args) -> Dict:
    """Make dataset and collator for supervised fine-tuning."""
    train_dataset = SupervisedDataset(tokenizer=tokenizer, data_path=data_args.data_path)
    data_collator = DataCollatorForSupervisedDataset(tokenizer=tokenizer)
    return dict(train_dataset=train_dataset, eval_dataset=None, data_collator=data_collator)


def train():
    parser = transformers.HfArgumentParser((ModelArguments, DataArguments, TrainingArguments))
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()

    model = transformers.AutoModelForCausalLM.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
    )

    tokenizer = transformers.AutoTokenizer.from_pretrained(
        model_args.model_name_or_path,
        cache_dir=training_args.cache_dir,
        model_max_length=training_args.model_max_length,
        padding_side="right",
        use_fast=False,
    )
    special_tokens_dict = dict()
    if tokenizer.pad_token is None:
        special_tokens_dict["pad_token"] = DEFAULT_PAD_TOKEN
    if tokenizer.eos_token is None:
        special_tokens_dict["eos_token"] = DEFAULT_EOS_TOKEN
    if tokenizer.bos_token is None:
        special_tokens_dict["bos_token"] = DEFAULT_BOS_TOKEN
    if tokenizer.unk_token is None:
        special_tokens_dict["unk_token"] = DEFAULT_UNK_TOKEN

    smart_tokenizer_and_embedding_resize(
        special_tokens_dict=special_tokens_dict,
        tokenizer=tokenizer,
        model=model,
    )

    data_module = make_supervised_data_module(tokenizer=tokenizer, data_args=data_args)
    trainer = Trainer(model=model, tokenizer=tokenizer, args=training_args, **data_module)
    trainer.train()
    trainer.save_state()
    trainer.save_model(output_dir=training_args.output_dir)


if __name__ == "__main__":
    train()



import argparse

import transformers
import torch


def init_eot_embedding_llama3(model_path, output_dir, special_tokens=["<|eot_id|>", "<|start_header_id|>", "<|end_header_id|>"], mean_cutoff=128000, dtype=torch.bfloat16):
    tokenizer = transformers.AutoTokenizer.from_pretrained(model_path)
    model = transformers.AutoModelForCausalLM.from_pretrained(model_path, low_cpu_mem_usage=True, torch_dtype=dtype)

    assert model.model.embed_tokens.weight.shape[0] >= mean_cutoff
    assert model.lm_head.weight.shape[0]            >= mean_cutoff

    with torch.no_grad():
        for token in special_tokens:
            token_id = tokenizer.convert_tokens_to_ids(token)

            print (f"Token {token} ID {token_id}")

            model.model.embed_tokens.weight[token_id] = torch.mean(model.model.embed_tokens.weight[:mean_cutoff].to(torch.float32), dim=0).to(dtype)
            model.lm_head.weight[token_id]            = torch.mean(model.lm_head.weight[:mean_cutoff].to(torch.float32), dim=0).to(dtype)

    # Save
    tokenizer.save_pretrained(output_dir)
    model.save_pretrained(output_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-path",
        help="Location of model, or HuggingFace repo ID",
    )
    parser.add_argument(
        "--output-dir",
        help="Location to write resulting model and tokenizer",
    )

    init_eot_embedding_llama3(**vars(parser.parse_args()))


if __name__ == "__main__":
    main()



import argparse
import os
import orjson

from glob import glob


def convert_to_evalplus(results_path: str, output_path: str):
    os.makedirs(output_path, exist_ok=True)

    for filename in glob(os.path.join(results_path, "*.json")):
        # read eval results
        with open(filename, "rb") as f:
            data = orjson.loads(f.read())

        # humaneval
        result = bytearray()
        for item in data:
            if item["task_type"] == "coding/humaneval":
                result.extend(orjson.dumps(item["answer"]))
                result.extend(b"\n")

        with open(os.path.join(output_path, os.path.splitext(os.path.basename(filename))[0] + ".jsonl"), "wb") as f:
            f.write(result)


def main():
    parser = argparse.ArgumentParser()

    # Input / output
    parser.add_argument("--results_path", type=str, default="ochat/evaluation/eval_results")
    parser.add_argument("--output_path",  type=str, default="ochat/evaluation/evalplus_codegen")
    args = parser.parse_args()

    convert_to_evalplus(**vars(args))


if __name__ == "__main__":
    main()



import re
import ast

from ochat.evaluation.grading.math_grader import grade_answer


def zs_agieval_match_answer(task_data, response):
    # AGIEval match first capital letter, following original paper implementation
    # https://github.com/microsoft/AGIEval/blob/main/src/post_process.py

    letter_set = {"A", "B", "C", "D", "E", "F"}
    for c in response:
        if c in letter_set:
            return True, c

    return False, ""


def zs_bbh_mc_orca_truthfulqa_orca_match_answer(task_data, response):
    # For BBH & TruthfulQA, match first option letter

    for c in response:
        if c in task_data["options"]:
            return True, c

    return False, ""


def fs_cothub_math_match_answer(task_data, response, max_length=256):
    def _last_boxed_only_string(string):
        idx = string.rfind("\\boxed")
        if idx < 0:
            idx = string.rfind("\\fbox")
            if idx < 0:
                return None

        i = idx
        left_brace_idx = None
        right_brace_idx = None
        num_left_braces_open = 0
        while i < len(string):
            if string[i] == "{":
                num_left_braces_open += 1
                if left_brace_idx is None:
                    left_brace_idx = i
            elif string[i] == "}":
                num_left_braces_open -= 1
                if num_left_braces_open == 0:
                    right_brace_idx = i
                    break

            i += 1
        
        if left_brace_idx is None or right_brace_idx is None:
            return None

        return string[left_brace_idx + 1: right_brace_idx].strip()

    # Match true answer
    ground_truth_answer = _last_boxed_only_string(task_data["_metadata"]["solution"])
    assert ground_truth_answer

    # Match model answer
    is_matched = False

    ans_line = response.split('The answer is')
    if len(ans_line) > 1:
        is_matched = True
        response = ans_line[-1].strip()
    else:
        ans_extracted = _last_boxed_only_string(response)
        if ans_extracted:
            is_matched = True
            response = ans_extracted

    # Grade
    response = response[:max_length]  # To avoid sympy taking too long
    return is_matched, grade_answer(response, ground_truth_answer)


def zs_gpqa_match_answer(task_data, response):
    # Expected to see answer field, otherwise return C.
    ans = response.split("The correct answer is")

    if len(ans) == 1:
        return False, "C"

    ans = ans[1]

    letter_set = {"A", "B", "C", "D"} 
    for c in ans:
        if c in letter_set:
            return True, c

    return False, "C"


def fs_cothub_bbh_match_answer(task_data, response):
    # CoT hub match answer for BBH
    # https://github.com/FranxYao/chain-of-thought-hub/blob/main/BBH/run_bbh_gpt_3.5_turbo.py

    ans_line = response.split('answer is ')

    # Expect to see 'answer is'. If not return whole string
    if len(ans_line) == 1:
        return False, response
    else:
        ans = ans_line[-1].strip()

    if task_data["options"]:
        # Multiple choice, find appearing letter
        options = ['(A)', '(B)', '(C)', '(D)', '(E)', '(F)', '(G)', '(H)', '(I)', '(J)', '(K)', '(L)', '(M)', '(N)', '(O)', '(P)', '(Q)', '(R)', '(S)', '(T)', '(U)', '(V)', '(W)', '(X)', '(Y)', '(Z)']

        for option in options:
            if option in ans:
                return True, option

        return False, ans
    else:
        # Free form, direct return
        if len(ans) and ans[-1] == '.':
            ans = ans[:-1]

        return True, ans


def fs_cothub_gsm8k_match_answer(task_data, response):
    # CoT hub match answer for GSM8k, match last numeric value
    # https://github.com/FranxYao/chain-of-thought-hub/blob/main/gsm8k/gpt3.5turbo_gsm8k_complex.ipynb

    pattern = '\d*\.?\d+'
    pred = re.findall(pattern, response)
    if len(pred) >= 1:
        return True, pred[-1]

    return False, response


def fs_cothub_mmlu_match_answer(task_data, response):
    ans_line = response.split('answer is')

    # Expect to see 'answer is'. If not return C
    if len(ans_line) == 1:
        return False, "(C)"
    else:
        ans = ans_line[-1].strip()
        
    options = ['(A)', '(B)', '(C)', '(D)']
    for option in options:
        if option in ans:
            return True, option

    return False, "(C)"


def coding_humaneval_match_answer(task_data, response):
    # Matching utilities
    def _function_exists(code, func_name):
        tree = ast.parse(code)
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef) and node.name == func_name:
                return True

        return False

    def _try_match(content, prefix, entrypoint):
        # All markdown code blocks, as well as raw
        code_blocks = [m[1] for m in re.findall(r"(\`{3}.*?\n+)([\s\S]*?)(\n+\`{3})", content)] \
                    + [content]

        for block in code_blocks:
            # Check syntax
            try:
                code_completion = prefix + block
                if _function_exists(code_completion, entrypoint):
                    return code_completion
            except SyntaxError:
                pass

    # Try match with include prefix
    humaneval_task = task_data["_metadata"]
    include_prefix = humaneval_task['prompt'].split('def')[0].strip() + "\n\n"

    result = _try_match(response, include_prefix, humaneval_task["entry_point"])
    if result: 
        return True, {"task_id": humaneval_task["task_id"], "completion": result}

    # If fail then match with function signature
    result = _try_match(response, humaneval_task["prompt"], humaneval_task["entry_point"])
    if result: 
        return True, {"task_id": humaneval_task["task_id"], "completion": result}

    return False, {"task_id": humaneval_task["task_id"], "completion": response}


MATCH_ANSWER_FUNCTION = {
    "zs/agieval": zs_agieval_match_answer,
    "zs/bbh_mc_orca": zs_bbh_mc_orca_truthfulqa_orca_match_answer,
    "zs/truthfulqa_orca": zs_bbh_mc_orca_truthfulqa_orca_match_answer,
    "zs/gpqa": zs_gpqa_match_answer,

    "fs_cothub/bbh": fs_cothub_bbh_match_answer,
    "fs_cothub/gsm8k": fs_cothub_gsm8k_match_answer,
    "fs_cothub/mmlu": fs_cothub_mmlu_match_answer,
    "fs_cothub/math": fs_cothub_math_match_answer,

    "coding/humaneval": coding_humaneval_match_answer
}



"""
Generate training data based on conversations

Usage: python -m ochat.data.generate_data --in-file sharegpt_gpt4.json --tokenizer-name HF_REPO_NAME --out-dir .
"""

from typing import Optional
from dataclasses import dataclass
import argparse
import json
import os
import random

import numpy as np
import transformers
from transformers.trainer_pt_utils import LabelSmoother
from ray.util.multiprocessing import Pool


@dataclass
class ModelDataConfig:
    name: str

    # Prompt
    system: str

    role_prefix: dict
    ai_role: str
    eot_token: str
    bos_token: Optional[str]

    # Tokenize
    max_tokens: int
    pad_token: int
    ignore_id: int


CONFIG = ModelDataConfig(
    name="OChat",

    # Prompt
    system="",

    role_prefix={
        "human": "Human: ",
        "gpt": "Assistant: "
    },
    ai_role="gpt",
    eot_token="<|end_of_turn|>",
    bos_token="<s>",

    # Tokenize
    max_tokens=8192,
    pad_token="<unk>",
    ignore_id=LabelSmoother.ignore_index
)


def generate_split(conversations: list, tokenizer: transformers.AutoTokenizer, split_name: str, out_dir: str):
    # Add prompt and tokenize conversation
    def _convert_single_conversation(c):
        tokens = []
        masks = []

        # begin of sentence (bos)
        if CONFIG.bos_token:
            t = tokenizer.convert_tokens_to_ids(CONFIG.bos_token)
            tokens.append(t)
            masks.append(False)

        # System
        if CONFIG.system:
            t = tokenizer(CONFIG.system, add_special_tokens=False) + [tokenizer.convert_tokens_to_ids(CONFIG.eot_token)]
            tokens.extend(t)
            masks.extend([False] * len(t))

        # Messages
        for message in c["items"]:
            # Message
            message_text = CONFIG.role_prefix[message["from"]] + message["value"]

            t = tokenizer(message_text, add_special_tokens=False) + [tokenizer.convert_tokens_to_ids(CONFIG.eot_token)]
            tokens.extend(t)
            masks.extend([message["from"] == CONFIG.ai_role] * len(t))

        return tokens, masks

    converted = Pool().map(_convert_single_conversation, conversations)

    # Pad and to numpy array
    pad_id = tokenizer.convert_tokens_to_ids(CONFIG.pad_token)

    all_input_ids = []
    all_labels = []
    all_attention_masks = []
    all_plain_texts = []
    for tokens, masks in converted:
        # Cut to length
        tokens = np.array(tokens[:CONFIG.max_tokens], np.int_)
        masks  = np.array(masks[:CONFIG.max_tokens], np.bool_)

        # Pad
        input_ids       = np.full(CONFIG.max_tokens, pad_id,           np.int_)
        labels          = np.full(CONFIG.max_tokens, CONFIG.ignore_id, np.int_)
        attention_masks = np.full(CONFIG.max_tokens, False,            np.bool_)

        length                   = len(tokens)

        input_ids[:length]       = tokens
        labels[:length]          = np.where(masks, tokens, CONFIG.ignore_id)
        attention_masks[:length] = True

        all_input_ids.append(input_ids)
        all_labels.append(labels)
        all_attention_masks.append(attention_masks)
        all_plain_texts.append(tokens)

    # Output training data
    np.savez(os.path.join(out_dir, f"ochat.{split_name}.npz"),
             # Arrays
             input_ids=np.vstack(all_input_ids),
             labels=np.vstack(all_labels),
             attention_masks=np.vstack(all_attention_masks))

    # Output plain texts
    all_plain_texts = tokenizer.decode(all_plain_texts)

    with open(os.path.join(out_dir, f"ochat.{split_name}.text.json"), "w") as f:
        json.dump(all_plain_texts, f)


def generate_dataset(seed, in_file, tokenizer_name, out_dir, eval_ratio):
    # Load tokenizer
    tokenizer = transformers.AutoTokenizer.from_pretrained(tokenizer_name, use_fast=False)

    # Load conversations
    with open(in_file, "r") as f:
        conversations = json.load(f)

    # Train-test split
    random.seed(seed)
    random.shuffle(conversations)
    eval_num = int(eval_ratio * len(conversations))

    train_conversations = conversations[eval_num:]
    eval_conversations  = conversations[:eval_num]

    generate_split(train_conversations, tokenizer, "train", out_dir)
    generate_split(eval_conversations, tokenizer, "eval", out_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--in-file", type=str, required=True)
    parser.add_argument("--tokenizer-name", type=str, required=True)
    parser.add_argument("--out-dir", type=str, default=".")
    parser.add_argument("--eval-ratio", type=float, default=0.01)
    args = parser.parse_args()

    generate_dataset(**vars(args))



from functools import partial

import torch
import transformers

from ochat.config.model_config import ModelConfig
from ochat.config.conversation_template import Message, Conversation, ConversationTemplate
import ochat.models


_GEMMA_IT_PREFIXES = {
    "user": "user",
    "assistant": "model"
}


def _v3_2_role_prefix(from_role, condition):
    return f"{condition} {from_role.title()}:".strip()


def _v3_6_role_prefix(from_role, condition, role_start_token, role_end_token):
    return role_start_token + f"{condition} {from_role.title()}".strip() + role_end_token


MODEL_CONFIG_MAP = {
    # OpenChat V3.6 (llama 3)
    "openchat_3.6": ModelConfig(
        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=True),  # Llama 3 only has fast tokenizer
        model_create_for_training=partial(ochat.models.LlamaForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),
        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=partial(_v3_6_role_prefix,
                                                          role_start_token="<|start_header_id|>",
                                                          role_end_token="<|end_header_id|>\n\n"),
                                      eot="<|eot_id|>",
                                      system_as_role=True,
                                      inference_condition="GPT4 Correct"),
        hf_chat_template="{{ bos_token }}{% for message in messages %}{% if message['role'] in ['user', 'assistant'] %}{% set content = '<|start_header_id|>GPT4 Correct ' + message['role'].title() + '<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' %}{% elif message['role'] == 'system' %}{% set content = '<|start_header_id|>System<|end_header_id|>\n\n' + message['content'] | trim + '<|eot_id|>' %}{% else %}{{ raise_exception('Only user, assistant and system roles are supported!') }}{% endif %}{{ content }}{% endfor %}{% if add_generation_prompt %}{{ '<|start_header_id|>GPT4 Correct Assistant<|end_header_id|>\n\n' }}{% endif %}",
    ),

    # OpenChat V3.2
    "openchat_v3.2": ModelConfig(
        # Model
        model_max_context=4096,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=False),
        model_create_for_training=partial(ochat.models.LlamaForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=_v3_2_role_prefix,
                                      eot="<|end_of_turn|>",
                                      inference_condition="GPT4")
    ),

    "openchat_v3.2_mistral": ModelConfig(
        serving_aliases=("openchat_3.5", ),

        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=True),
        model_create_for_training=partial(ochat.models.MistralForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=_v3_2_role_prefix,
                                      eot="<|end_of_turn|>",
                                      inference_condition="GPT4 Correct"),
        hf_chat_template="{{ bos_token }}{% for message in messages %}{% if message['role'] in ['user', 'assistant'] %}{% set content = 'GPT4 Correct ' + message['role'].title() + ': ' + message['content'] + '<|end_of_turn|>' %}{% elif message['role'] == 'system' %}{% set content = message['content'] + '<|end_of_turn|>' %}{% else %}{{ raise_exception('Only user, assistant and system roles are supported!') }}{% endif %}{{ content }}{% endfor %}{% if add_generation_prompt %}{{ 'GPT4 Correct Assistant:' }}{% endif %}",    ),

    "openchat_v3.2_gemma_new": ModelConfig(
        serving_aliases=("openchat_3.5_gemma_new", ),

        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=True),
        model_create_for_training=partial(ochat.models.GemmaForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=_v3_2_role_prefix,
                                      eot="<end_of_turn>",
                                      inference_condition="GPT4 Correct"),
        hf_chat_template="{{ bos_token }}{% for message in messages %}{% if message['role'] in ['user', 'assistant'] %}{% set content = 'GPT4 Correct ' + message['role'].title() + ': ' + message['content'] + '<end_of_turn>' %}{% elif message['role'] == 'system' %}{% set content = message['content'] + '<end_of_turn>' %}{% else %}{{ raise_exception('Only user, assistant and system roles are supported!') }}{% endif %}{{ content }}{% endfor %}{% if add_generation_prompt %}{{ 'GPT4 Correct Assistant:' }}{% endif %}",
    ),

    ### Other models
    "chatml_8192": ModelConfig(
        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=True),
        model_create_for_training=lambda x: None,

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=lambda from_role, condition: f"<|im_start|>{from_role}\n",
                                      eot="<|im_end|>",
                                      inference_condition="")
    ),
    "zephyr_mistral": ModelConfig(
        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=False),
        model_create_for_training=partial(ochat.models.MistralForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=lambda from_role, condition: f"<|{from_role}|>\n",
                                      eot="</s>",
                                      inference_condition="")
    ),
    "gemma_it": ModelConfig(
        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=False),
        model_create_for_training=partial(ochat.models.GemmaForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=lambda from_role, condition: f"<start_of_turn>{_GEMMA_IT_PREFIXES[from_role]}\n",
                                      eot="<end_of_turn>",
                                      inference_condition="")
    ),
    "llama3_instruct": ModelConfig(
        # Model
        model_max_context=8192,
        model_tokenizer_create=partial(transformers.AutoTokenizer.from_pretrained, use_fast=True),  # Llama 3 only has fast tokenizer
        model_create_for_training=partial(ochat.models.LlamaForCausalLM.from_pretrained,
                                          low_cpu_mem_usage=True,
                                          torch_dtype=torch.bfloat16),

        # Conversation Template
        conversation_template=partial(ConversationTemplate,
                                      role_prefix=lambda from_role, condition: f"<|start_header_id|>{from_role}<|end_header_id|>\n\n",
                                      eot="<|eot_id|>",
                                      inference_condition="")
    ),
}
